/***********************************************************************************************************************
 *
 * blueBill Core - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://java.net/hg/bluebill~core-src
 *
 **********************************************************************************************************************/
package it.tidalwave.observation.simple;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.io.Serializable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import it.tidalwave.util.EmptyInitializer;
import it.tidalwave.util.Id;
import it.tidalwave.util.Initializer;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.role.spi.DefaultDisplayable;
import it.tidalwave.role.spi.DefaultIdentifiable;
import org.openide.util.Lookup;
import it.tidalwave.netbeans.util.AsLookupSupport;
import it.tidalwave.observation.Finder;
import it.tidalwave.observation.Location;
import it.tidalwave.observation.Observable;
import it.tidalwave.observation.Observation;
import it.tidalwave.observation.Observation.Builder;
import it.tidalwave.observation.ObservationItem;
import it.tidalwave.observation.ObservationSet;
import it.tidalwave.observation.Observer;
import it.tidalwave.observation.Source;
import it.tidalwave.observation.spi.ObservationSetSupport;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class SimpleObservationSet extends ObservationSetSupport implements ObservationSet, Serializable
  {
    private static final Logger log = LoggerFactory.getLogger(SimpleObservationSet.class);
    private static final long serialVersionUID = 2356635648437593487L;

    private final static Map<Class<?>, Factory<?>> FINDER_FACTORY_MAP_BY_CLASS = new HashMap<Class<?>, Factory<?>>();

    private final static Map<Class<?>, FactoryWithInitializer<?>> FACTORY_MAP_BY_CLASS = new HashMap<Class<?>, FactoryWithInitializer<?>>();

    private final List<Observation> observations = new ArrayList<Observation>();

    private final AsLookupSupport asSupport;
    
    @CheckForNull
    private transient Map<String, Object> cache;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    interface Factory<T>
      {
        @Nonnull
        public T create (@Nonnull SimpleObservationSet observationSet);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    interface FactoryWithInitializer<T>
      {
        @Nonnull
        public T create (@Nonnull String displayName, @Nonnull Initializer<T> initializer);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static final Factory<Finder<Observation>> OBSERVATION_FINDER_FACTORY = new Factory<Finder<Observation>>()
      {
        @Nonnull
        public Finder<Observation> create (final @Nonnull SimpleObservationSet observationSet)
          {
            return new SimpleFinderSupport<Observation>(observationSet.observations);
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static final Factory<Finder<Observable>> OBSERVABLE_FINDER_FACTORY = new Factory<Finder<Observable>>()
      {
        @Nonnull
        public Finder<Observable> create (final @Nonnull SimpleObservationSet observationSet)
          {
            final Set<Observable> observables = new HashSet<Observable>();

            for (final Observation observation : observationSet.observations)
              {
                for (final ObservationItem observationItem : observation.findObservationItems().results())
                  {
                    observables.add(observationItem.getObservable());
                  }
              }

            return new SimpleFinderSupport<Observable>(observables);
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static final Factory<Finder<Observer>> OBSERVER_FINDER_FACTORY = new Factory<Finder<Observer>>()
      {
        @Nonnull
        public Finder<Observer> create (final @Nonnull SimpleObservationSet observationSet)
          {
            final Set<Observer> observers = new HashSet<Observer>();

            for (final Observation observation : observationSet.observations)
              {
                for (final Observer observer : observation.findObservers().results())
                  {
                    observers.add(observer);
                  }
              }

            return new SimpleFinderSupport<Observer>(observers);
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static final Factory<Finder<Location>> LOCATION_FINDER_FACTORY = new Factory<Finder<Location>>()
      {
        @Nonnull
        public Finder<Location> create (final @Nonnull SimpleObservationSet observationSet)
          {
            final Set<Location> locations = new HashSet<Location>();

            for (final Observation observation : observationSet.observations)
              {
                locations.add(observation.getLocation());
              }

            return new SimpleFinderSupport<Location>(locations);
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static final FactoryWithInitializer<Observable> OBSERVABLE_FACTORY = new FactoryWithInitializer<Observable>()
      {
        @Nonnull
        public Observable create (final @Nonnull String displayName,
                                  final @Nonnull Initializer<Observable> initializer)
          {
            return initializer.initialize(new SimpleObservable(new DefaultDisplayable(displayName, "SimpleObservable")));
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static final FactoryWithInitializer<Observer> OBSERVER_FACTORY = new FactoryWithInitializer<Observer>()
      {
        @Nonnull
        public Observer create (final @Nonnull String displayName,
                                final @Nonnull Initializer<Observer> initializer)
          {
            return initializer.initialize(new SimpleObserver(new DefaultDisplayable(displayName, "SimpleObserver")));
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static final FactoryWithInitializer<Location> LOCATION_FACTORY = new FactoryWithInitializer<Location>()
      {
        @Nonnull
        public Location create (final @Nonnull String displayName,
                                final @Nonnull Initializer<Location> initializer)
          {
            return initializer.initialize(new SimpleLocation(new DefaultDisplayable(displayName, "SimpleLocation")));
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static final FactoryWithInitializer<Source> SOURCE_FACTORY = new FactoryWithInitializer<Source>()
      {
        @Nonnull
        public Source create (final @Nonnull String displayName,
                              final @Nonnull Initializer<Source> initializer)
          {
            return initializer.initialize(new SimpleSource(new DefaultDisplayable(displayName, "SimpleSource")));
          }
      };


    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    static
      {
        FINDER_FACTORY_MAP_BY_CLASS.put(Observation.class, OBSERVATION_FINDER_FACTORY);
        FINDER_FACTORY_MAP_BY_CLASS.put(Observable.class, OBSERVABLE_FINDER_FACTORY);
        FINDER_FACTORY_MAP_BY_CLASS.put(Observer.class, OBSERVER_FINDER_FACTORY);
        FINDER_FACTORY_MAP_BY_CLASS.put(Location.class, LOCATION_FINDER_FACTORY);

        FACTORY_MAP_BY_CLASS.put(Observable.class, OBSERVABLE_FACTORY);
        FACTORY_MAP_BY_CLASS.put(Observer.class, OBSERVER_FACTORY);
        FACTORY_MAP_BY_CLASS.put(Location.class, LOCATION_FACTORY);
        FACTORY_MAP_BY_CLASS.put(Source.class, SOURCE_FACTORY);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public SimpleObservationSet (final @Nonnull Object ... instanceCapabilities)
      {
        asSupport = new AsLookupSupport(this, instanceCapabilities);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public <T> T findOrCreate (final @Nonnull Class<T> clazz,
                               final @Nonnull String displayName)
      {
        return findOrCreate(clazz, displayName, new EmptyInitializer<T>());
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public synchronized <T> T findOrCreate (final @Nonnull Class<T> clazz,
                                            final @Nonnull String displayName,
                                            final @Nonnull Initializer<T> initializer)
      {
        final String key = clazz.getName() + ":" + displayName;

        if (cache == null)
          {
            cache = new HashMap<String, Object>();
          }

        T result = (T)cache.get(key);

        if (result == null)
          {
            final FactoryWithInitializer<T> factory = (FactoryWithInitializer<T>)FACTORY_MAP_BY_CLASS.get(clazz);

            if (factory == null)
              {
                throw new IllegalArgumentException("Not supported: " + clazz);
              }

            result = factory.create(displayName, initializer);
            cache.put( key, result);
          }

        return result;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public <T> T findById (final @Nonnull Class<T> clazz, final @Nonnull Id id)
      throws NotFoundException
      {
        throw new UnsupportedOperationException("Not supported yet.");
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public <T> Finder<T> find (final @Nonnull Class<T> clazz)
      {
        final Factory<Finder<T>> finderFactory = (Factory<Finder<T>>)FINDER_FACTORY_MAP_BY_CLASS.get(clazz);

        if (finderFactory == null)
          {
            throw new IllegalArgumentException("Not supported: " + clazz);
          }

        return finderFactory.create(this);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public Builder createObservation()
      {
        return new SimpleObservationBuilder(this);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public Source createSource (String uri, String displayName)
      {
        return new SimpleSource(new DefaultIdentifiable(new Id(uri)), new DefaultDisplayable("...", displayName));
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void clear()
      {
        observations.clear();
        fireNotifyCleared();
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public final <T> T as (final @Nonnull Class<T> type, final @Nonnull NotFoundBehaviour<T> nfb)
      {
        return asSupport.as(type, nfb);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public final <T> T as (final @Nonnull Class<T> type)
      {
        return asSupport.as(type);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Lookup getLookup() 
      {
        return asSupport.getLookup();
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void addObservation (final @Nonnull SimpleObservation observation)
      {
        observations.add(observation);
        fireObservationAdded(observation);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void replaceObservation (final @Nonnull Observation originalObservation,
                                       final @Nonnull SimpleObservation observation)
      {
        final int index = observations.indexOf(originalObservation);

        if (index < 0)
          {
            throw new IllegalStateException("The original observation has been deleted in the meantime " + originalObservation);
          }

        observations.set(index, observation);
        fireObservationChanged(observation, originalObservation);
      }
  }
